#include "SparseFilter.h"
#include "WeightMap.h"
#include "Utils.h"
#include <algorithm>
#include <cmath>
#include <cassert>

using std::max;

namespace sparse_filter {

namespace {

inline int SymmetricCoordinate(int pos, int size) {
  if (pos < 0) pos = -pos - 1;
  if (pos >= 2 * size) pos %= 2 * size;
  if (pos >= size) pos = 2 * size - 1 - pos;
  return pos;
}

void PadSymmetric(const Image &src, int pad_before, int pad_after, Image *dst) {
  int pad_total = pad_before + pad_after;
  dst->Init(src.rows() + pad_total, src.columns() + pad_total, src.channels());
  for (int row = 0; row < dst->rows(); ++row) {
    for (int col = 0; col < dst->columns(); ++col) {
      for (int chan = 0; chan < dst->channels(); ++chan) {
        dst->val(col, row, chan) = src.val(SymmetricCoordinate(col - pad_before, src.columns()),
                                           SymmetricCoordinate(row - pad_before, src.rows()),
                                           chan);
      }
    }
  }
}

void ExtractPatch(const Image &src, int pr, int pc, Image *dst) {
  // src is padded, so (pr, pc) becomes the upper left pixel
  for (int row = 0; row < dst->rows(); ++row) {
    for (int col = 0; col < dst->columns(); ++col) {
      for (int chan = 0; chan < dst->channels(); ++chan) {
        dst->val(col, row, chan) = src.val(pc + col, pr + row, chan);
      }
    }
  }
}

//void FastExtractPatch(const Image &src, int pr, int pc, Image *dst) {
//  // src is padded, so (pr, pc) becomes the upper left pixel
//  int i = 0, j = (pr * src.columns() + pc) * src.channels();
//  for (int row = 0; row < dst->rows(); ++row, j += (src.columns() - dst->columns()) * src.channels()) {
//    for (int el = 0; el < dst->columns() * dst->channels(); ++el, ++i, ++j) {
//      dst->val(i) = src.val(j);
//    }
//  }
//}

    void BilateralWeight(const Image &g, Image *k, int r, float sigma_d2,
                         float sigma_r2) {
        for (int row = 0; row < g.rows(); ++row) {
            for (int col = 0; col < g.columns(); ++col) {
                float x = 0.f;
                for (int chan = 0; chan < g.channels(); ++chan) {
                    float y = g.val(col, row, chan) - g.val(r, r, chan);
                    x += y * y;
                }
                x /= 2 * sigma_r2;
                x += ((row - r) * (row - r) + (col - r) * (col - r)) / (2 * sigma_d2);
                k->val(col, row) = utils::fastexp(-x);
            }
        }
    }

void ComputeAverage(const Image &y, const Image &k, float yt[]) {
  float total_weight = 0.f;
  for (int chan = 0; chan < y.channels(); ++chan) {
    yt[chan] = 0.f;
  }

  for (int row = 0; row < y.rows(); ++row) {
    for (int col = 0; col < y.columns(); ++col) {
      total_weight += k.val(col, row);
      for (int chan = 0; chan < y.channels(); ++chan) {
        yt[chan] += y.val(col, row, chan) * k.val(col, row);
      }
    }
  }

  for (int chan = 0; chan < y.channels(); ++chan) {
    yt[chan] /= total_weight;
  }
}

void AggregateValue(const float *yt, const Image &k, int start_row, int start_col, Image *output) {
  // icol and irow are the "internal" indexes (with respect to the patch).
  // col and row are the indexes with respect to the entire image
  for (int row = max(0, start_row), irow = row - start_row; irow < k.rows() && row < output->rows(); ++row, ++irow) {
    for (int col = max(0, start_col), icol = col - start_col; icol < k.columns() && col < output->columns(); ++col, ++icol) {
      for (int chan = 0; chan < output->channels(); ++chan) {
        output->val(col, row, chan) += yt[chan] * k.val(icol, irow);
      }
    }
  }
}

void SquarePatch(Image *k) {
  for (int row = 0; row < k->rows(); ++row) {
    for (int col = 0; col < k->columns(); ++col) {
      k->val(col, row) *= k->val(col, row);
    }
  }
}

void Normalize(const WeightMap &weights, Image *image) {
  for (int row = 0; row < image->rows(); ++row) {
    for (int col = 0; col < image->columns(); ++col) {
      for (int chan = 0; chan < image->channels(); ++chan) {
        image->val(col, row, chan) /= weights.val(col, row);
      }
    }
  }
}

}  // namespace

    
void SparseFilter(const Image& source, float sigma_d2, float sigma_r2, int r,
                  float threshold, Image* output) {
  // initializations and internal variables
  output->Init(source.rows(), source.columns(), source.channels());
  output->Reset();
  WeightMap agg_weights(source.rows(), source.columns());
  int pr, pc;  // coordinates of the central pixel
  int s = 2 * r + 1;
  Image y(s, s, source.channels());  // extracted patch
  float yt[source.channels()];  // average value
  Image k(s, s);  // weight function
  Image source_pad;
  // image padding
  PadSymmetric(source, r, r, &source_pad);
  // main loop
  while (agg_weights.Minimum() < threshold) {
    agg_weights.FindMinimum(&pr, &pc);
    ExtractPatch(source_pad, pr, pc, &y);
    BilateralWeight(y, &k, r, sigma_d2, sigma_r2);
    ComputeAverage(y, k, yt);
    AggregateValue(yt, k, pr - r, pc - r, output);
//    SquarePatch(&k);
    agg_weights.IncreaseWeights(k, pr - r, pc - r);
  }
  Normalize(agg_weights, output);
}
    

    /*
     * Bilateral Filter WITHOUT Regression.
     *      LA FONCTION DE CHARLES
     *      la seule faite par lui
     *      dans tout ce paquet.
     *      (avec la suivante)
     */
    
    void BilateralFilter(const Image& source, float sigma_d2, float sigma_r2, int r, Image* output) {
        
        // ------------------------------------------------------------------------------ //
        // Pseudo-code du filtre bilatéral avec régression : (mais ici on fait pas la reg.)
        /*
         1. En dehors de la boucle :
         ---------------------------
         
         1.a) Construction du noyau spatial
         
         Gaussienne spatialle sur un rayon r*r et de variance sigma_d2
         
         1.b) Gestion des bords
         
         Recopie en miroir de l'image d'entrée sur un nombre r de pixels
         de chaque côté de l'image
         
         1.c) Allocation de mémoire pour les variables...
         
         
         2. Dans la boucle : POUR CHAQUE PIXEL (h,w)
         -------------------------------------------
         
         2.a) Construction du noyan gaussien en intensité
         
         Voisinage dans r*r moins la valeur en (h,w) (valeur centrale)
         auquel on applique la fonction gaussienne de paramètre sigma_r2 (variance)
         
         2.b) Calcul des poids spatiaux du bilatéral
         
         Multiplication point à point des poids spatiaux et d'intensité.
         
         2.c) Calcul des valeurs "SIGMA" nécessaires :
         
         Voir mon document sur l'accélération du filtre bilatéral avec régression,
         on utilise ici la solution analytique et on ne calcule que le paramètre c
         du plan de régression (seul nécessaire), avec la méthode de Cramer.
         
         2.d) Résolution du système
         
         Comme je le dis juste au-dessus.
         
         2.e) L'image de sortie prend le paramètre c à la position (h,w)
         
         FIN DE L'ALGO
         */
        // ------------------------------------------------------------------------------ //
        
        
        
        // Initialization of the output image
        output->Init(source.rows(), source.columns(), source.channels());
        output->Reset();
        
        // Some variables
        int pr, pc;                         // coordinates of the central pixel
        int s = 2 * r + 1;                  // patch size
        Image y(s, s, source.channels());   // extracted patch
        Image k(s, s);                      // weights
        Image source_pad;                   // source image with padding
        
        // Some other variables
        float total_weight = 0.f;
        float yt[y.channels()];
        int row, col, chan;
        
        // Image padding
        PadSymmetric(source, r, r, &source_pad); // +++ Utiliser source_pad dans la suite
        
        
        
        // Loop over the whole image
        for(pr=0;pr<source.rows();pr++){
            for(pc=0;pc<source.columns();pc++){
                
                // Patch centered in (pr,pc)
                ExtractPatch(source_pad, pr, pc, &y);
                
                // Computation of the bilateral weights
                BilateralWeight(y, &k, r, sigma_d2, sigma_r2);
                
                // Compute the filtered value at (pr,pc)
                for (chan = 0; chan < y.channels(); ++chan) {
                    yt[chan] = 0.f;
                }
                total_weight = 0.f; // total_weight reset for each new window
                for (row = 0; row < y.rows(); ++row) {
                    for (col = 0; col < y.columns(); ++col) {
                        total_weight += k.val(col, row);
                        for (chan = 0; chan < y.channels(); ++chan) {
                            yt[chan] += y.val(col, row, chan) * k.val(col, row);
                        }
                    }
                }
                // Normalization
                for (chan = 0; chan < y.channels(); ++chan) {
                    output->val(pc,pr,chan) = yt[chan]/total_weight;
                }
            }
        }
    }
    
    
    
    /*
     * Bilateral Filter WITH Regression.
     *      LA FONCTION DE CHARLES
     *      la seule faite par lui
     *      dans tout ce paquet.
     *      (avec la précédente)
     */
    
    void BilateralFilterReg(const Image& source, float sigma_d2, float sigma_r2, int r, Image* output) {
        
        // ------------------------------------------------------------------------------ //
        // Pseudo-code du filtre bilatéral avec régression :
        /*
         1. En dehors de la boucle :
         ---------------------------
         
         1.a) Construction du noyau spatial
         
         Gaussienne spatialle sur un rayon r*r et de variance sigma_d2
         
         1.b) Gestion des bords
         
         Recopie en miroir de l'image d'entrée sur un nombre r de pixels
         de chaque côté de l'image
         
         1.c) Allocation de mémoire pour les variables...
         
         
         2. Dans la boucle : POUR CHAQUE PIXEL (h,w)
         -------------------------------------------
         
         2.a) Construction du noyan gaussien en intensité
         
         Voisinage dans r*r moins la valeur en (h,w) (valeur centrale)
         auquel on applique la fonction gaussienne de paramètre sigma_r2 (variance)
         
         2.b) Calcul des poids spatiaux du bilatéral
         
         Multiplication point à point des poids spatiaux et d'intensité.
         
         2.c) Calcul des valeurs "SIGMA" nécessaires :
         
         Voir mon document sur l'accélération du filtre bilatéral avec régression,
         on utilise ici la solution analytique et on ne calcule que le paramètre c
         du plan de régression (seul nécessaire), avec la méthode de Cramer.
         
         2.d) Résolution du système
         
         Comme je le dis juste au-dessus.
         
         2.e) L'image de sortie prend le paramètre c à la position (h,w)
         
         FIN DE L'ALGO
         */
        // ------------------------------------------------------------------------------ //
        
        
        
        // Initialization of the output image
        output->Init(source.rows(), source.columns(), source.channels());
        output->Reset();
        
        // Some variables
        int pr, pc;                         // coordinates of the central pixel
        int s = 2 * r + 1;                  // patch size
        Image y(s, s, source.channels());   // extracted patch
        Image k(s, s);                      // weights
        Image source_pad;                   // source image with padding
        Image x1(s,s);                      // x-axis (horizontal)
        Image x2(s,s);                      // y-axis (vertical)
        int i1,i2,j;                        // counters
        
        // Some other variables
        float total_weight = 0.f;
        float y_sums[3][y.channels()];
        float x_sums[6];
        float c1=0.f, c2=0.f;
        int row, col, chan;
        
        // Image padding
        PadSymmetric(source, r, r, &source_pad); // +++ Utiliser source_pad dans la suite
        
        // Generation of x1 and x2
        for(i1=0;i1<s;i1++){
            for(i2=0;i2<s;i2++){
                x1.val(i1,i2) = (float)(i1-r);
                x2.val(i1,i2) = (float)(i2-r);
            }
        }
        
        // Loop over the whole image
        for(pr=0;pr<source.rows();pr++){
            for(pc=0;pc<source.columns();pc++){
                
                // Patch centered in (pr,pc)
                ExtractPatch(source_pad, pr, pc, &y);
                
                // Computation of the bilateral weights
                BilateralWeight(y, &k, r, sigma_d2, sigma_r2);
                
                // -=- Compute the filtered value at (pr,pc) -=-
                
                // Initialization of y_sums
                for (chan = 0; chan < y.channels(); ++chan) {
                    for(j=0;j<3;j++) y_sums[j][chan] = 0.f;
                }
                
                // Initialisation x_sums for each new window
                for(j=0;j<6;j++) x_sums[j] = 0.f;
                
                // Loop over the patch y
                for (row = 0; row < y.rows(); ++row) {
                    for (col = 0; col < y.columns(); ++col) {
                        
                        x_sums[0] += k.val(col, row);
                        x_sums[1] += k.val(col, row) * x1.val(col,row);
                        x_sums[2] += k.val(col, row) * x2.val(col,row);
                        x_sums[3] += k.val(col, row) * x1.val(col,row) * x1.val(col,row);
                        x_sums[4] += k.val(col, row) * x1.val(col,row) * x2.val(col,row);
                        x_sums[5] += k.val(col, row) * x2.val(col,row) * x2.val(col,row);
                        
                        for (chan = 0; chan < y.channels(); ++chan) {
                            y_sums[0][chan] += y.val(col, row, chan) * k.val(col, row);
                            y_sums[1][chan] += y.val(col, row, chan) * k.val(col, row) * x1.val(col,row);
                            y_sums[2][chan] += y.val(col, row, chan) * k.val(col, row) * x2.val(col,row);
                        }
                    }
                }
                // -=- Computation of the plan parameter "c" -=-
                
                // Matrice 2 determinant
                c2 = 0.f;
                c2 =
                + x_sums[3] * (x_sums[5]*x_sums[0] - x_sums[2]*x_sums[2])
                - x_sums[4] * (x_sums[4]*x_sums[0] - x_sums[1]*x_sums[2])
                + x_sums[1] * (x_sums[4]*x_sums[2] - x_sums[1]*x_sums[5]);
                
                
                for (chan = 0; chan < y.channels(); ++chan) {
                    // Matrice 1 determinant
                    c1 = 0.f;
                    c1 =
                    + x_sums[3]       * (x_sums[5]*y_sums[0][chan] - x_sums[2]*y_sums[2][chan])
                    - x_sums[4]       * (x_sums[4]*y_sums[0][chan] - x_sums[1]*y_sums[2][chan])
                    + y_sums[1][chan] * (x_sums[4]*x_sums[2]       - x_sums[1]*x_sums[5]      );
                    
                    output->val(pc,pr,chan) = c1/c2;
                }
            }
        }
        
    }


    void difference(const Image& source, const Image& output, Image* diff, float coeffmult) {

        // Initialization of the output image
        diff->Init(source.rows(), source.columns(), source.channels());
        diff->Reset();

        for (int row = 0; row < source.rows(); ++row) {
            for (int col = 0; col < source.columns(); ++col) {
              for (int chan = 0; chan < source.channels(); ++chan) {
                diff->val(col, row, chan) = 128 + coeffmult*(source.val(col, row, chan) - output.val(col, row, chan));
            }
        }
    }

    }


}  // namespace da3d
